@@ -103,6 +103,7 @@ struct _BluetoothClientPrivate {
 	DBusGProxy *manager;
 	GtkTreeStore *store;
 	GtkTreeRowReference *default_adapter;
+	GSList *horrible_workaround_for_leaked_ifaces;
 };
 
 enum {
@@ -266,6 +267,7 @@ device_services_changed (DBusGProxy *iface, const char *property,
 static GHashTable *
 device_list_nodes (DBusGProxy *device, BluetoothClient *client, gboolean connect_signal)
 {
+	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
 	GHashTable *table;
 	guint i;
 
@@ -333,6 +335,8 @@ device_list_nodes (DBusGProxy *device, BluetoothClient *client, gboolean connect
 				dbus_g_proxy_connect_signal(iface, "PropertyChanged",
 							    G_CALLBACK(device_services_changed), client, NULL);
 			}
+
+			priv->horrible_workaround_for_leaked_ifaces = g_slist_append (priv->horrible_workaround_for_leaked_ifaces, iface);
 		}
 	}
 
@@ -1021,9 +1025,41 @@ bluetooth_client_set_property (GObject        *object,
 	}
 }
 
+static gboolean
+disconnect_from_proxy_helper (GtkTreeModel *model,
+			      GtkTreePath  *path,
+			      GtkTreeIter  *iter,
+			      gpointer      data)
+{
+	BluetoothClient *client = data;
+	DBusGProxy *proxy;
+
+	gtk_tree_model_get(model, iter,
+			   BLUETOOTH_COLUMN_PROXY, &proxy, -1);
+
+	/* adapters */
+	g_signal_handlers_disconnect_by_func(proxy,
+				adapter_changed, client);
+	g_signal_handlers_disconnect_by_func(proxy,
+				device_created, client);
+	g_signal_handlers_disconnect_by_func(proxy,
+				device_removed, client);
+	g_signal_handlers_disconnect_by_func(proxy,
+				device_found, client);
+
+	/* devices */
+	g_signal_handlers_disconnect_by_func(proxy,
+				device_changed, client);
+
+	g_object_unref(proxy);
+
+	return FALSE;
+}
+
 static void bluetooth_client_finalize(GObject *client)
 {
 	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
+	GSList *l;
 
 	DBG("client %p", client);
 
@@ -1041,11 +1077,22 @@ static void bluetooth_client_finalize(GObject *client)
 					default_adapter_changed, client);
 	g_object_unref(priv->manager);
 
+	gtk_tree_model_foreach (GTK_TREE_MODEL(priv->store), disconnect_from_proxy_helper, client);
+
 	g_object_unref(priv->store);
 
 	if (priv->default_adapter)
 		gtk_tree_row_reference_free (priv->default_adapter);
 
+	for (l = priv->horrible_workaround_for_leaked_ifaces; l != NULL; l = l->next) {
+		DBusGProxy *iface = l->data;
+
+		g_signal_handlers_disconnect_by_func(iface,
+					device_services_changed, client);
+		g_object_unref (iface);
+	}
+	g_slist_free (priv->horrible_workaround_for_leaked_ifaces);
+
 	G_OBJECT_CLASS(bluetooth_client_parent_class)->finalize(client);
 }
 
@@ -1140,7 +1187,7 @@ GtkTreeModel *bluetooth_client_get_model (BluetoothClient *client)
  * @data: user data to pass to gtk_tree_model_filter_set_visible_func()
  * @destroy: a destroy function for gtk_tree_model_filter_set_visible_func()
  *
- * Returns a #GtkTreeModel of devices filtered using the @func, @data and @destroy arguments to pass to gtk_tree_model_filter_set_visible_func().
+ * Returns a #GtkTreeModelFilter of devices filtered using the @func, @data and @destroy arguments to pass to gtk_tree_model_filter_set_visible_func().
  *
  * Return value: (transfer full): a #GtkTreeModel object.
  **/
@@ -1187,7 +1234,7 @@ static gboolean adapter_filter(GtkTreeModel *model,
  * bluetooth_client_get_adapter_model:
  * @client: a #BluetoothClient object
  *
- * Returns a filtered #GtkTreeModel with only adapters present.
+ * Returns a #GtkTreeModelFilter with only adapters present.
  *
  * Return value: (transfer full): a #GtkTreeModel object.
  **/
@@ -1204,7 +1251,10 @@ GtkTreeModel *bluetooth_client_get_adapter_model (BluetoothClient *client)
  * @client: a #BluetoothClient object
  * @adapter: a #DBusGProxy of the adapter object, or %NULL to get the default adapter.
  *
- * Returns a filtered #GtkTreeModel with only devices belonging to the selected adapter listed. Note that the model will follow a specific adapter, and will not follow the default-adapter when %NULL is passed.
+ * Returns a #GtkTreeModelFilter with only devices belonging to the @adapter listed.
+ * Note that the model will follow a specific adapter, and will not follow the default adapter.
+ * Also note that due to the way #GtkTreeModelFilter works, you will probably want to
+ * monitor signals on the "child-model" #GtkTreeModel to monitor for changes.
  *
  * Return value: (transfer full): a #GtkTreeModel object.
  **/
